package net.asukaze.cassava.io.util;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.util.ArrayList;
import java.util.List;

import net.asukaze.cassava.core.model.CassavaTableModel;
import net.asukaze.cassava.core.model.DataFormat;
import net.asukaze.cassava.core.model.ICassavaTableModel;

public class CsvReader {
	public enum NextType {
		HAS_MORE_CELL, HAS_MORE_ROW, END_OF_FILE
	}

	public enum DelimiterType {
		NONE, WEAK, STRONG
	}

	private static final String LINE_SEPARATOR = "\n"; //$NON-NLS-1$
	private static final String EMPTY = ""; //$NON-NLS-1$

	private DataFormat dataFormat;
	private BufferedReader reader;
	private String line;
	private List<String> cells;

	public CsvReader(DataFormat dataFormat, Reader reader) {
		this.dataFormat = dataFormat;
		this.reader = new BufferedReader(reader);
		this.line = null;
		this.cells = null;
	}

	public void close() throws IOException {
		reader.close();
	}

	@Override
	protected void finalize() throws IOException {
		close();
	}

	public NextType getNextType() throws IOException {
		if (cells != null && cells.size() > 0) {
			return NextType.HAS_MORE_CELL;
		}
		if (line == null) {
			line = reader.readLine();
		}
		if (line != null) {
			return NextType.HAS_MORE_ROW;
		} else {
			return NextType.END_OF_FILE;
		}
	}

	public List<String> splitCsv(String line, BufferedReader reader)
			throws IOException {
		boolean quoted = false;
		DelimiterType delimiterType = DelimiterType.STRONG;
		List<String> cells = new ArrayList<String>();
		StringBuilder cell = new StringBuilder();

		while (line != null) {
			for (int i = 0; i < line.length(); i++) {
				char ch = line.charAt(i);
				if (!quoted) {
					if (dataFormat.isStrongDelimiter(ch)) {
						if (delimiterType != DelimiterType.WEAK) {
							cells.add(cell.toString());
							cell = new StringBuilder();
						}
						delimiterType = DelimiterType.STRONG;
					} else if (dataFormat.isWeakDelimiter(ch)) {
						if (delimiterType == DelimiterType.NONE) {
							cells.add(cell.toString());
							cell = new StringBuilder();
							delimiterType = DelimiterType.WEAK;
						}
					} else if (dataFormat.isUseQuote()
							&& dataFormat.isQuoteChar(ch)
							&& delimiterType != DelimiterType.NONE) {
						quoted = true;
						delimiterType = DelimiterType.NONE;
					} else {
						cell.append(ch);
						delimiterType = DelimiterType.NONE;
					}
				} else { // Quoted
					if (dataFormat.isUseQuote() && dataFormat.isQuoteChar(ch)) {
						if (i < line.length() - 1 && line.charAt(i + 1) == ch) {
							cell.append(ch);
							i++;
						} else {
							cells.add(cell.toString());
							cell = new StringBuilder();
							delimiterType = DelimiterType.WEAK;
							quoted = false;
						}
					} else {
						cell.append(ch);
						delimiterType = DelimiterType.NONE;
					}
				}
			}
			if (quoted) {
				cell.append(LINE_SEPARATOR);
				line = reader.readLine();
				// loop
			} else {
				line = null;
				// break;
			}
		}

		if (cell.length() > 0) {
			cells.add(cell.toString());
		} else if (delimiterType == DelimiterType.STRONG) {
			cells.add(EMPTY);
		}
		return cells;
	}

	public String next() throws IOException {
		if (cells != null && cells.size() > 0) {
			String value = cells.get(0);
			cells = cells.subList(1, cells.size());
			if (cells.size() == 0) {
				cells = null;
			}
			return value;
		}
		if (line == null) {
			line = reader.readLine();
		}
		if (line != null) {
			cells = splitCsv(line, reader);
			line = null;
			return next();
		} else {
			return null;
		}
	}

	public ICassavaTableModel readTable(ICassavaTableModel model)
			throws IOException {
		model.clear();
		int y = -1;
		int x = 0;
		while (true) {
			NextType type = getNextType();
			if (type == NextType.END_OF_FILE) {
				break;
			} else if (type == NextType.HAS_MORE_ROW) {
				y++;
				x = 0;
			}
			model.setValueAt(next(), y, x);
			x++;
		}
		model.setModified(false);
		return model;
	}

	public ICassavaTableModel readTable() throws IOException {
		return readTable(new CassavaTableModel());
	}
}
