package net.asukaze.cassava.macro;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import net.asukaze.cassava.core.model.ICassavaTableModel;
import net.asukaze.cassava.core.model.ICassavaViewModel;
import net.asukaze.cassava.macro.api.CoreFunction;
import net.asukaze.cassava.macro.parser.ASTAPIFunctionCall;
import net.asukaze.cassava.macro.parser.ASTAdd;
import net.asukaze.cassava.macro.parser.ASTArgument;
import net.asukaze.cassava.macro.parser.ASTCellRef;
import net.asukaze.cassava.macro.parser.ASTDivision;
import net.asukaze.cassava.macro.parser.ASTEquals;
import net.asukaze.cassava.macro.parser.ASTExpr01;
import net.asukaze.cassava.macro.parser.ASTExtFunctionCall;
import net.asukaze.cassava.macro.parser.ASTFor;
import net.asukaze.cassava.macro.parser.ASTFor2;
import net.asukaze.cassava.macro.parser.ASTFunctionCall;
import net.asukaze.cassava.macro.parser.ASTFunctionDef;
import net.asukaze.cassava.macro.parser.ASTGreaterThan;
import net.asukaze.cassava.macro.parser.ASTGreaterThanEqual;
import net.asukaze.cassava.macro.parser.ASTIf;
import net.asukaze.cassava.macro.parser.ASTLeftDec;
import net.asukaze.cassava.macro.parser.ASTLeftInc;
import net.asukaze.cassava.macro.parser.ASTLessThan;
import net.asukaze.cassava.macro.parser.ASTLessThanEqual;
import net.asukaze.cassava.macro.parser.ASTLogicalAnd;
import net.asukaze.cassava.macro.parser.ASTLogicalOr;
import net.asukaze.cassava.macro.parser.ASTMod;
import net.asukaze.cassava.macro.parser.ASTMulti;
import net.asukaze.cassava.macro.parser.ASTNotEquals;
import net.asukaze.cassava.macro.parser.ASTNotValue;
import net.asukaze.cassava.macro.parser.ASTNumber;
import net.asukaze.cassava.macro.parser.ASTReturn;
import net.asukaze.cassava.macro.parser.ASTRightDec;
import net.asukaze.cassava.macro.parser.ASTRightInc;
import net.asukaze.cassava.macro.parser.ASTSbst;
import net.asukaze.cassava.macro.parser.ASTScript;
import net.asukaze.cassava.macro.parser.ASTSpecialVarRef;
import net.asukaze.cassava.macro.parser.ASTStatement;
import net.asukaze.cassava.macro.parser.ASTString;
import net.asukaze.cassava.macro.parser.ASTSub;
import net.asukaze.cassava.macro.parser.ASTVarRef;
import net.asukaze.cassava.macro.parser.ASTWhile;
import net.asukaze.cassava.macro.parser.CMSParser;
import net.asukaze.cassava.macro.parser.CMSParserVisitor;
import net.asukaze.cassava.macro.parser.Node;
import net.asukaze.cassava.macro.parser.ParseException;
import net.asukaze.cassava.macro.parser.SimpleNode;
import net.asukaze.cassava.macro.value.CellReference;
import net.asukaze.cassava.macro.value.Reference;
import net.asukaze.cassava.macro.value.SpecialVarReference;
import net.asukaze.cassava.macro.value.VarReference;

public class Macro {
	private CMSParserVisitor visitor;
	private Environment environment;
	private Map<String, ASTScript> scriptsCache;
	private List<APIFunctionHandler> apiFunctionHandlers;
	private ScriptHandler scriptHandler;

	public Macro() {
		this.visitor = new Visitor();
		this.environment = new Environment();
		this.scriptsCache = new HashMap<String, ASTScript>();
		this.apiFunctionHandlers = new ArrayList<APIFunctionHandler>();
		this.registerAPIFunctionHandler(new CoreFunction.Handler());
		this.scriptHandler = new DefaultScriptHandler();
	}

	public void setModel(ICassavaTableModel model) {
		environment.setModel(model);
	}

	public void setViewModel(ICassavaViewModel viewModel) {
		environment.setViewModel(viewModel);
	}

	public Environment getEnvironment() {
		return environment;
	}

	public Object getVar(String name) {
		return environment.getVar(name);
	}

	public void setScriptHandler(ScriptHandler scriptHandler) {
		this.scriptHandler = scriptHandler;
	}

	public void setAPIFunctionHandlers(
			List<APIFunctionHandler> apiFunctionHandlers) {
		this.apiFunctionHandlers = apiFunctionHandlers;
	}

	public List<APIFunctionHandler> getAPIFunctionHandlers() {
		return apiFunctionHandlers;
	}

	public void registerAPIFunctionHandler(APIFunctionHandler apiFunctionHandler) {
		this.apiFunctionHandlers.add(apiFunctionHandler);
	}

	public void clearAPIFunctionHandlers() {
		this.apiFunctionHandlers = new ArrayList<APIFunctionHandler>();
	}

	public void close() throws IOException {
		for (APIFunctionHandler handler : apiFunctionHandlers) {
			handler.close();
		}
	}

	public Object dereference(Object obj) {
		return environment.dereference(obj);
	}

	public Double toDouble(Object obj) {
		return environment.toDouble(obj);
	}

	public String toString(Object obj) {
		return environment.toString(obj);
	}

	public int toInt(Object obj) {
		return environment.toInt(obj);
	}

	public boolean isNumber(Object obj) {
		obj = dereference(obj);
		return (obj == null || obj instanceof Double || obj instanceof Integer);
	}

	public ASTScript getScriptTree(String fileid) {
		if (scriptsCache.containsKey(fileid)) {
			return scriptsCache.get(fileid);
		}
		if (scriptHandler != null) {
			InputStream inputStream = scriptHandler.open(fileid);
			try {
				if (inputStream != null) {
					CMSParser parser = new CMSParser(inputStream);
					ASTScript node = parser.Script();
					scriptsCache.put(fileid, node);
					return node;
				}
			} catch (ParseException ex) {
				ex.printStackTrace();
			} finally {
				try {
					if (inputStream != null) {
						inputStream.close();
					}
				} catch (IOException ex) {
					ex.printStackTrace();
				}
			}
		}
		return null;
	}

	public void run(InputStream inputStream) throws ParseException, IOException {
		try {
			run(new InputStreamReader(inputStream));
		} finally {
			inputStream.close();
		}
	}

	public void run(Reader reader) throws ParseException, IOException {
		try {
			CMSParser parser = new CMSParser(reader);
			ASTScript node = parser.Script();
			node.jjtAccept(visitor, null);
		} catch (MacroException ex) {
			Throwable cause = ex.getCause();
			if (cause instanceof IOException) {
				throw (IOException) cause;
			} else {
				throw ex;
			}
		} finally {
			reader.close();
		}
	}

	public Object runExpr(InputStream inputStream) throws ParseException,
			IOException {
		try {
			return runExpr(new InputStreamReader(inputStream));
		} finally {
			inputStream.close();
		}
	}

	public Object runExpr(Reader reader) throws ParseException, IOException {
		try {
			CMSParser parser = new CMSParser(reader);
			ASTExpr01 node = parser.Expr01();
			return node.jjtAccept(visitor, null);
		} catch (MacroException ex) {
			Throwable cause = ex.getCause();
			if (cause instanceof IOException) {
				throw (IOException) cause;
			} else {
				throw ex;
			}
		} finally {
			reader.close();
		}
	}

	public void pushEnvironment(String fileid) {
		environment.push(fileid);
	}

	public void popEnvironment() {
		environment.pop();
	}

	public Object accept(Node node) {
		return node.jjtAccept(visitor, null);
	}

	public class Visitor implements CMSParserVisitor {
		public Object visit(SimpleNode node, Object data) {
			throw new RuntimeException("SimpleNode"); //$NON-NLS-1$
		}

		public Object visit(ASTScript node, Object data) {
			node.childrenAccept(this, null);
			return null;
		}

		public Object visit(ASTStatement node, Object data) {
			int n = node.jjtGetNumChildren();
			if (n > 0) {
				for (int i = 0; i < n - 1; i++) {
					node.jjtGetChild(i).jjtAccept(this, null);
				}
				return node.jjtGetChild(n - 1).jjtAccept(this, null);
			} else {
				return null;
			}
		}

		public Object visit(ASTExpr01 node, Object data) {
			int n = node.jjtGetNumChildren();
			if (n > 0) {
				return node.jjtGetChild(0).jjtAccept(this, null);
			} else {
				return data;
			}
		}

		private FunctionCall createFunctionCall(SimpleNode node) {
			String name = node.nodeValue;
			int n = node.jjtGetNumChildren();
			Object[] args = new Object[n];
			for (int i = 0; i < n; i++) {
				Object ref = node.jjtGetChild(i).jjtAccept(this, null);
				args[i] = dereference(ref);
			}
			return new FunctionCall(name, args);
		}

		private Object runAPIFunction(FunctionCall call) throws IOException {
			for (APIFunctionHandler handler : apiFunctionHandlers) {
				Object result = handler.runAPIFunction(call.getName(), call
						.getArgs(), Macro.this);
				if (!(result instanceof FunctionNotFoundException)) {
					return result;
				}
			}
			return new FunctionNotFoundException(call);
		}

		public Object visit(ASTFunctionCall node, Object data) {
			try {
				FunctionCall call = createFunctionCall(node);
				Object result = UserFunction.run(call, Macro.this, false);
				if (result instanceof FunctionNotFoundException) {
					result = runAPIFunction(call);
					if (result instanceof FunctionNotFoundException) {
						throw (FunctionNotFoundException) result;
					}
				}
				return result;
			} catch (IOException ex) {
				throw new MacroException(ex);
			}
		}

		public Object visit(ASTAPIFunctionCall node, Object data) {
			try {
				FunctionCall call = createFunctionCall(node);
				Object result = runAPIFunction(call);
				if (result instanceof FunctionNotFoundException) {
					result = UserFunction.run(call, Macro.this, false);
					if (result instanceof FunctionNotFoundException) {
						throw (FunctionNotFoundException) result;
					}
				}
				return result;
			} catch (IOException ex) {
				throw new MacroException(ex);
			}
		}

		public Object visit(ASTExtFunctionCall node, Object data) {
			FunctionCall call = createFunctionCall(node);
			Object result = UserFunction.run(call, Macro.this, true);
			if (result instanceof FunctionNotFoundException) {
				throw (FunctionNotFoundException) result;
			}
			return result;
		}

		public Object visit(ASTReturn node, Object data) {
			int n = node.jjtGetNumChildren();
			Object value = null;
			if (n > 0) {
				value = dereference(node.jjtGetChild(0).jjtAccept(this, null));
			}
			throw new ReturnException(value);
		}

		public Object visit(ASTVarRef node, Object data) {
			String name = node.nodeValue;
			return new VarReference(name);
		}

		public Object visit(ASTSpecialVarRef node, Object data) {
			String name = node.nodeValue;
			return new SpecialVarReference(name);
		}

		public Object visit(ASTCellRef node, Object data) {
			int x = toInt(node.jjtGetChild(0).jjtAccept(this, null));
			int y = toInt(node.jjtGetChild(1).jjtAccept(this, null));
			return new CellReference(x, y);
		}

		public Object visit(ASTSbst node, Object data) {
			Object left = node.jjtGetChild(0).jjtAccept(this, null);
			Object right = node.jjtGetChild(1).jjtAccept(this, null);
			if (left instanceof Reference) {
				Reference var = ((Reference) left);
				Object value = dereference(right);
				var.set(value, environment);
			} else {
				throw new MacroException(Messages
						.getString(Messages.NON_LVALUE));
			}
			return null;
		}

		public Object visit(ASTLogicalOr node, Object data) {
			int x = toInt(node.jjtGetChild(0).jjtAccept(this, null));
			int y = toInt(node.jjtGetChild(1).jjtAccept(this, null));
			boolean result = (x != 0 || y != 0);
			return Integer.valueOf(result ? 1 : 0);
		}

		public Object visit(ASTLogicalAnd node, Object data) {
			int x = toInt(node.jjtGetChild(0).jjtAccept(this, null));
			int y = toInt(node.jjtGetChild(1).jjtAccept(this, null));
			boolean result = (x != 0 && y != 0);
			return Integer.valueOf(result ? 1 : 0);
		}

		public Object visit(ASTEquals node, Object data) {
			String left = Macro.this.toString(node.jjtGetChild(0).jjtAccept(
					this, null));
			String right = Macro.this.toString(node.jjtGetChild(1).jjtAccept(
					this, null));
			return Integer.valueOf(left.equals(right) ? 1 : 0);
		}

		public Object visit(ASTNotEquals node, Object data) {
			String left = Macro.this.toString(node.jjtGetChild(0).jjtAccept(
					this, null));
			String right = Macro.this.toString(node.jjtGetChild(1).jjtAccept(
					this, null));
			return Integer.valueOf((!left.equals(right)) ? 1 : 0);
		}

		public Object visit(ASTLessThan node, Object data) {
			Double left = toDouble(node.jjtGetChild(0).jjtAccept(this, null));
			Double right = toDouble(node.jjtGetChild(1).jjtAccept(this, null));
			return Integer.valueOf((left < right) ? 1 : 0);
		}

		public Object visit(ASTLessThanEqual node, Object data) {
			Double left = toDouble(node.jjtGetChild(0).jjtAccept(this, null));
			Double right = toDouble(node.jjtGetChild(1).jjtAccept(this, null));
			return Integer.valueOf((left <= right) ? 1 : 0);
		}

		public Object visit(ASTGreaterThan node, Object data) {
			Double left = toDouble(node.jjtGetChild(0).jjtAccept(this, null));
			Double right = toDouble(node.jjtGetChild(1).jjtAccept(this, null));
			return Integer.valueOf((left > right) ? 1 : 0);
		}

		public Object visit(ASTGreaterThanEqual node, Object data) {
			Double left = toDouble(node.jjtGetChild(0).jjtAccept(this, null));
			Double right = toDouble(node.jjtGetChild(1).jjtAccept(this, null));
			return Integer.valueOf((left >= right) ? 1 : 0);
		}

		public Object visit(ASTAdd node, Object data) {
			Object left = node.jjtGetChild(0).jjtAccept(this, null);
			Object right = node.jjtGetChild(1).jjtAccept(this, null);
			if (isNumber(left) && isNumber(right)) {
				return toDouble(left) + toDouble(right);
			}
			return Macro.this.toString(left) + Macro.this.toString(right);
		}

		public Object visit(ASTSub node, Object data) {
			Double left = toDouble(node.jjtGetChild(0).jjtAccept(this, null));
			Double right = toDouble(node.jjtGetChild(1).jjtAccept(this, null));
			return left - right;
		}

		public Object visit(ASTMulti node, Object data) {
			Double left = toDouble(node.jjtGetChild(0).jjtAccept(this, null));
			Double right = toDouble(node.jjtGetChild(1).jjtAccept(this, null));
			return left * right;
		}

		public Object visit(ASTDivision node, Object data) {
			Double left = toDouble(node.jjtGetChild(0).jjtAccept(this, null));
			Double right = toDouble(node.jjtGetChild(1).jjtAccept(this, null));
			return left / right;
		}

		public Object visit(ASTMod node, Object data) {
			int left = toInt(node.jjtGetChild(0).jjtAccept(this, null));
			int right = toInt(node.jjtGetChild(1).jjtAccept(this, null));
			return Integer.valueOf(left % right);
		}

		public Object visit(ASTRightInc node, Object data) {
			String name = node.nodeValue;
			Object obj = getEnvironment().getVar(name);
			getEnvironment().putVar(name, toDouble(obj) + 1.0);
			return obj;
		}

		public Object visit(ASTRightDec node, Object data) {
			String name = node.nodeValue;
			Object obj = getEnvironment().getVar(name);
			getEnvironment().putVar(name, toDouble(obj) - 1.0);
			return obj;
		}

		public Object visit(ASTLeftInc node, Object data) {
			String name = node.nodeValue;
			Double value = toDouble(getEnvironment().getVar(name)) + 1.0;
			getEnvironment().putVar(name, value);
			return value;
		}

		public Object visit(ASTLeftDec node, Object data) {
			String name = node.nodeValue;
			Double value = toDouble(getEnvironment().getVar(name)) - 1.0;
			getEnvironment().putVar(name, value);
			return value;
		}

		public Object visit(ASTNotValue node, Object data) {
			int value = toInt(node.jjtGetChild(0).jjtAccept(this, null));
			return Integer.valueOf(value == 0 ? 1 : 0);
		}

		private boolean condition(Node node) {
			return (toInt(node.jjtAccept(this, 1)) != 0);
		}

		public Object visit(ASTIf node, Object data) {
			int n = node.jjtGetNumChildren();
			if (condition(node.jjtGetChild(0))) {
				node.jjtGetChild(1).jjtAccept(this, null);
			} else if (n >= 3) {
				node.jjtGetChild(2).jjtAccept(this, null);
			}
			return null;
		}

		public Object visit(ASTWhile node, Object data) {
			Node condNode = node.jjtGetChild(0);
			Node blockNode = node.jjtGetChild(1);
			while (condition(condNode)) {
				blockNode.jjtAccept(this, null);
			}
			return null;
		}

		public Object visit(ASTFor node, Object data) {
			Node initNode = node.jjtGetChild(0);
			Node condNode = node.jjtGetChild(1);
			Node nextNode = node.jjtGetChild(2);
			Node blockNode = node.jjtGetChild(3);

			for (initNode.jjtAccept(this, null); condition(condNode); nextNode
					.jjtAccept(this, null)) {
				blockNode.jjtAccept(this, null);
			}
			return null;
		}

		public Object visit(ASTFor2 node, Object data) {
			String name = node.nodeValue;
			Node maxNode = node.jjtGetChild(0);
			Node blockNode = node.jjtGetChild(1);

			for (int i = 1; i <= toInt(maxNode.jjtAccept(this, null)); i++) {
				getEnvironment().putVar(name, Integer.valueOf(i));
				blockNode.jjtAccept(this, null);
			}
			return null;
		}

		int hex(char ch) {
			if (ch >= '0' && ch <= '9') {
				return (ch - '0');
			} else if (ch >= 'A' && ch <= 'F') {
				return (ch - 'A') + 10;
			} else if (ch >= 'a' && ch <= 'f') {
				return (ch - 'a') + 10;
			}
			return 0;
		}

		public Object visit(ASTString node, Object data) {
			StringBuilder value = new StringBuilder();
			String raw = node.nodeValue;
			for (int i = 0; i < raw.length(); i++) {
				char c = raw.charAt(i);
				if (c == '\\') {
					char n = raw.charAt(++i);
					if (n == 'n') {
						value.append('\n');
					} else if (n == 'r') {
						value.append('\r');
					} else if (n == 't') {
						value.append('\t');
					} else if (n == 'x') {
						int code = hex(raw.charAt(++i)) * 16
								+ hex(raw.charAt(++i));
						value.append((char) code);
					} else {
						value.append(n);
					}
				} else {
					value.append(c);
				}
			}
			return value.toString();
		}

		public Object visit(ASTNumber node, Object data) {
			return toDouble(node.nodeValue);
		}

		public Object visit(ASTFunctionDef node, Object data) {
			String fileid = environment.getCurrentFileId();
			int argCount = node.jjtGetNumChildren() - 1;
			UserFunction userFunction = new UserFunction(node, fileid);
			environment.putFunction(fileid, node.nodeValue, argCount,
					userFunction);
			return null;
		}

		public Object visit(ASTArgument node, Object data) {
			return node.nodeValue;
		}
	}
}
