package net.asukaze.cassava.macro.api;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;

import javax.swing.JOptionPane;

import net.asukaze.cassava.macro.APIFunctionHandler;
import net.asukaze.cassava.macro.FunctionNotFoundException;
import net.asukaze.cassava.macro.Macro;
import net.asukaze.cassava.macro.Messages;

public class CUIMessageBoxFunctionHandler implements APIFunctionHandler {
	private static final String TITLE_PREFIX = "[ "; //$NON-NLS-1$
	private static final String TITLE_POSTFIX = " ]"; //$NON-NLS-1$
	private static final String DEF_PREFIX = "("; //$NON-NLS-1$
	private static final String DEF_POSTFIX = ")"; //$NON-NLS-1$

	private BufferedReader consoleReader;
	private PrintStream consoleWriter;

	public CUIMessageBoxFunctionHandler(BufferedReader consoleReader,
			PrintStream consoleWriter) {
		this.consoleReader = consoleReader;
		this.consoleWriter = consoleWriter;
	}

	public CUIMessageBoxFunctionHandler(InputStream in, PrintStream out) {
		this.consoleReader = new BufferedReader(new InputStreamReader(in));
		this.consoleWriter = out;
	}

	public Object runAPIFunction(String name, Object[] args, Macro macro)
			throws IOException {
		int n = args.length;
		if ("MessageBox".equals(name) && n <= 3) { //$NON-NLS-1$
			return showMessageBox(args, macro);
		} else if ("InputBox".equals(name) && n <= 3) { //$NON-NLS-1$
			return showInputBox(args, macro);
		}
		return new FunctionNotFoundException(name, n);
	}

	private Integer showMessageBox(Object[] args, Macro macro)
			throws IOException {
		if (args.length == 0) {
			return showMessageBox(Messages.getString(Messages.BREAK_POINT),
					null, JOptionPane.DEFAULT_OPTION);
		} else if (args.length == 1) {
			return showMessageBox(macro.toString(args[0]), null,
					JOptionPane.DEFAULT_OPTION);
		} else if (args.length == 2) {
			return showMessageBox(macro.toString(args[0]), macro
					.toString(args[1]), JOptionPane.DEFAULT_OPTION);
		} else {
			return showMessageBox(macro.toString(args[0]), macro
					.toString(args[1]), macro.toInt(args[2]));
		}
	}

	private Integer showMessageBox(String message, String title, int optionType)
			throws IOException {
		if (title != null) {
			consoleWriter.print(TITLE_PREFIX);
			consoleWriter.print(title);
			consoleWriter.println(TITLE_POSTFIX);
		}
		if (message != null) {
			consoleWriter.println(message);
		}
		String prompt;
		boolean canYes = false;
		boolean canNo = false;
		boolean canOk = false;
		boolean canCancel = false;
		if (optionType == JOptionPane.YES_NO_OPTION) {
			prompt = "Yes(Y), No(N) ? ";
			canYes = true;
			canNo = true;
		} else if (optionType == JOptionPane.YES_NO_CANCEL_OPTION) {
			prompt = "Yes(Y), No(N), Cancel(C) ? ";
			canYes = true;
			canNo = true;
			canCancel = true;
		} else if (optionType == JOptionPane.OK_CANCEL_OPTION) {
			prompt = "OK(Enter), Cancel(C) ? ";
			canOk = true;
			canCancel = true;
		} else {
			return JOptionPane.OK_OPTION;
		}
		while (true) {
			consoleWriter.print(prompt);
			String line = consoleReader.readLine();
			if (line == null) {
				return null;
			} else if (line.length() == 0) {
				if (canOk) {
					return JOptionPane.OK_OPTION;
				}
			} else if (line.charAt(0) == 'Y' || line.charAt(0) == 'y') {
				if (canYes) {
					return JOptionPane.YES_OPTION;
				}
			} else if (line.charAt(0) == 'N' || line.charAt(0) == 'n') {
				if (canNo) {
					return JOptionPane.NO_OPTION;
				}
			} else if (line.charAt(0) == 'O' || line.charAt(0) == 'o') {
				if (canOk) {
					return JOptionPane.OK_OPTION;
				}
			} else if (line.charAt(0) == 'C' || line.charAt(0) == 'c') {
				if (canCancel) {
					return JOptionPane.CANCEL_OPTION;
				}
			}
		}
	}

	private String showInputBox(Object[] args, Macro macro) throws IOException {
		if (args.length == 0) {
			return showInputBox(null, null, null);
		} else if (args.length == 1) {
			return showInputBox(macro.toString(args[0]), null, null);
		} else if (args.length == 2) {
			return showInputBox(macro.toString(args[0]), null, macro
					.toString(args[1]));
		} else {
			return showInputBox(macro.toString(args[0]), macro
					.toString(args[1]), macro.toString(args[2]));
		}
	}

	private String showInputBox(String message, String title, String def)
			throws IOException {
		if (title != null) {
			consoleWriter.print(TITLE_PREFIX);
			consoleWriter.print(title);
			consoleWriter.println(TITLE_POSTFIX);
		}
		if (message != null) {
			consoleWriter.println(message);
		}
		String prompt;
		if (def != null && def.length() > 0) {
			prompt = DEF_PREFIX + def + DEF_POSTFIX + "? ";
		} else {
			prompt = "? ";
		}
		consoleWriter.print(prompt);
		String line = consoleReader.readLine();
		if (line == null || line.length() == 0) {
			return def;
		} else {
			return line;
		}
	}

	public void close() {
	}
}
