package net.asukaze.cassava.macro.api;

import java.io.IOException;
import java.security.SecureRandom;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Random;

import net.asukaze.cassava.core.model.ICassavaTableModel;
import net.asukaze.cassava.core.model.ICassavaViewModel;
import net.asukaze.cassava.macro.APIFunctionHandler;
import net.asukaze.cassava.macro.Environment;
import net.asukaze.cassava.macro.FunctionNotFoundException;
import net.asukaze.cassava.macro.Macro;
import net.asukaze.cassava.macro.MacroException;
import net.asukaze.cassava.macro.Messages;
import net.asukaze.cassava.macro.value.CellReference;
import net.asukaze.cassava.macro.value.MidReference;
import net.asukaze.cassava.macro.value.Reference;
import net.asukaze.cassava.macro.value.Void;

public class CoreFunction {
	private static final String EMPTY = ""; //$NON-NLS-1$

	private String name;
	private Object[] args;
	private Macro macro;
	private static Random rand = null;

	public CoreFunction(String name, Object[] args, Macro macro) {
		this.name = name;
		this.args = args.clone();
		this.macro = macro;
	}

	public static class Handler implements APIFunctionHandler {
		public Object runAPIFunction(String name, Object[] args, Macro macro)
				throws IOException {
			return new CoreFunction(name, args, macro).run();
		}

		public void close() {
		}
	}

	private Double toDouble(Object obj) {
		return macro.toDouble(obj);
	}

	private String toString(Object obj) {
		return macro.toString(obj);
	}

	private int toInt(Object obj) {
		return macro.toInt(obj);
	}

	public Object run() throws IOException {
		int n = args.length;
		if ("InsertRow".equals(name) && n <= 1) { //$NON-NLS-1$
			return insertRow(args);
		} else if ("DeleteRow".equals(name) && n <= 1) { //$NON-NLS-1$
			return deleteRow(args);
		} else if ("InsertCol".equals(name) && n <= 1) { //$NON-NLS-1$
			return insertCol(args);
		} else if ("DeleteCol".equals(name) && n <= 1) { //$NON-NLS-1$
			return deleteCol(args);
		} else if ("SetRowHeight".equals(name) && n >= 1 && n <= 2) { //$NON-NLS-1$
			return setRowHeight(args);
		} else if ("SetColWidth".equals(name) && n >= 1 && n <= 2) { //$NON-NLS-1$
			return setColWidth(args);
		} else if ("max".equals(name) && n >= 1) { //$NON-NLS-1$
			return max(args);
		} else if ("min".equals(name) && n >= 1) { //$NON-NLS-1$
			return min(args);
		} else if ("mid".equals(name) && n >= 2 && n <= 3) { //$NON-NLS-1$
			return mid(args);
		} else if ("GetRowHeight".equals(name) && n <= 1) { //$NON-NLS-1$
			return getRowHeight(args);
		} else if ("GetColWidth".equals(name) && n <= 1) { //$NON-NLS-1$
			return getColWidth(args);
		} else if (n == 0) {
			if ("GetFileName".equals(name)) { //$NON-NLS-1$
				return getFileName();
			} else if ("GetFilePath".equals(name)) { //$NON-NLS-1$
				return getFilePath();
			} else if ("GetYear".equals(name)) { //$NON-NLS-1$
				return getYear();
			} else if ("GetMonth".equals(name)) { //$NON-NLS-1$
				return getMonth();
			} else if ("GetDate".equals(name)) { //$NON-NLS-1$
				return getDate();
			} else if ("GetHours".equals(name)) { //$NON-NLS-1$
				return getHours();
			} else if ("GetMinutes".equals(name)) { //$NON-NLS-1$
				return getMinutes();
			} else if ("GetSeconds".equals(name)) { //$NON-NLS-1$
				return getSeconds();
			}
		} else if (n == 1) {
			if ("int".equals(name)) { //$NON-NLS-1$
				return toInt(args[0]);
			} else if ("double".equals(name)) { //$NON-NLS-1$
				return toDouble(args[0]);
			} else if ("str".equals(name)) { //$NON-NLS-1$
				return toString(args[0]);
			} else if ("len".equals(name)) { //$NON-NLS-1$
				return toString(args[0]).length();
			} else if ("lenB".equals(name)) { //$NON-NLS-1$
				return toString(args[0]).getBytes().length;
			} else if ("random".equals(name)) { //$NON-NLS-1$
				return random(toInt(args[0]));
			}
		} else if (n == 2) {
			if ("cell".equals(name)) { //$NON-NLS-1$
				return cell(toInt(args[0]), toInt(args[1]));
			} else if ("move".equals(name)) { //$NON-NLS-1$
				return move(toInt(args[0]), toInt(args[1]));
			} else if ("moveto".equals(name)) { //$NON-NLS-1$
				return moveto(toInt(args[0]), toInt(args[1]));
			} else if ("swap".equals(name)) { //$NON-NLS-1$
				return swap(args[0], args[1]);
			} else if ("left".equals(name)) { //$NON-NLS-1$
				return left(toString(args[0]), toInt(args[1]));
			} else if ("right".equals(name)) { //$NON-NLS-1$
				return right(toString(args[0]), toInt(args[1]));
			} else if ("pos".equals(name)) { //$NON-NLS-1$
				return pos(toString(args[0]), toString(args[1]));
			} else if ("posB".equals(name)) { //$NON-NLS-1$
				return posB(toString(args[0]), toString(args[1]));
			}
		} else if (n == 3) {
			if ("replace".equals(name)) { //$NON-NLS-1$
				return replace(toString(args[0]), toString(args[1]),
						toString(args[2]));
			}
		} else if (n == 4) {
			if ("sum".equals(name)) { //$NON-NLS-1$
				return sum(toInt(args[0]), toInt(args[1]), toInt(args[2]),
						toInt(args[3]));
			} else if ("avr".equals(name)) { //$NON-NLS-1$
				return avr(toInt(args[0]), toInt(args[1]), toInt(args[2]),
						toInt(args[3]));
			}
		}
		return new FunctionNotFoundException(name, n);
	}

	private CellReference cell(int x, int y) {
		return new CellReference(x, y);
	}

	private Void move(int vx, int vy) {
		ICassavaViewModel viewModel = macro.getEnvironment().getViewModel();
		int row = viewModel.getCurrentRow();
		int col = viewModel.getCurrentCol();
		viewModel.setCurrentRowCol(row + vy, col + vx);
		return null;
	}

	private Void moveto(int x, int y) {
		ICassavaViewModel viewModel = macro.getEnvironment().getViewModel();
		viewModel.setCurrentRowCol(y, x);
		return null;
	}

	private Void swap(Object a, Object b) {
		if (a instanceof Reference && b instanceof Reference) {
			Environment environment = macro.getEnvironment();
			Reference aRef = (Reference) a;
			Reference bRef = (Reference) b;
			Object aValue = environment.dereference(a);
			Object bValue = environment.dereference(b);
			aRef.set(bValue, environment);
			bRef.set(aValue, environment);
			return null;
		} else {
			throw new MacroException(Messages
					.getString(Messages.SWAP_NOT_LEFT_VALUE));
		}
	}

	private Void insertRow(Object[] args) {
		ICassavaTableModel model = macro.getEnvironment().getModel();
		if (args.length == 0) {
			model.insertRow(model.getRowCount());
		} else {
			model.insertRow(toInt(args[0])-1);
		}
		return null;
	}

	private Void deleteRow(Object[] args) {
		ICassavaTableModel model = macro.getEnvironment().getModel();
		if (args.length == 0) {
			model.deleteRow(model.getRowCount()-1);
		} else {
			model.deleteRow(toInt(args[0])-1);
		}
		return null;
	}

	private Void insertCol(Object[] args) {
		ICassavaTableModel model = macro.getEnvironment().getModel();
		if (args.length == 0) {
			model.insertColumn(model.getColumnCount());
		} else {
			model.insertColumn(toInt(args[0])-1);
		}
		return null;
	}

	private Void deleteCol(Object[] args) {
		ICassavaTableModel model = macro.getEnvironment().getModel();
		if (args.length == 0) {
			model.deleteColumn(model.getColumnCount()-1);
		} else {
			model.deleteColumn(toInt(args[0])-1);
		}
		return null;
	}

	private Void setRowHeight(Object[] args) {
		ICassavaViewModel viewModel = macro.getEnvironment().getViewModel();
		if (args.length == 1) {
			viewModel.setDefaultRowHeight(toInt(args[0]));
			viewModel.setRowHeights(new ArrayList<Integer>());
		} else {
			viewModel.setRowHeight(toInt(args[0]), toInt(args[1]));
		}
		return null;
	}

	private Void setColWidth(Object[] args) {
		ICassavaViewModel viewModel = macro.getEnvironment().getViewModel();
		if (args.length == 1) {
			viewModel.setDefaultColWidth(toInt(args[0]));
			viewModel.setColWidths(new ArrayList<Integer>());
		} else {
			viewModel.setColWidth(toInt(args[0]), toInt(args[1]));
		}
		return null;
	}

	private Double max(Object[] args) {
		double maxValue = toDouble(args[0]);
		for (int i = 1; i < args.length; i++) {
			maxValue = Math.max(maxValue, toDouble(args[i]));
		}
		return maxValue;
	}

	private Double min(Object[] args) {
		double minValue = toDouble(args[0]);
		for (int i = 1; i < args.length; i++) {
			minValue = Math.min(minValue, toDouble(args[i]));
		}
		return minValue;
	}

	private String left(String str, int x) {
		if (x > 0 && x < str.length()) {
			return str.substring(0, x);
		} else if (x <= 0) {
			return EMPTY;
		} else {
			return str;
		}
	}

	private String right(String str, int x) {
		int endIndex = str.length();
		int beginIndex = endIndex - x;
		if (beginIndex >= 0) {
			return str.substring(beginIndex, endIndex);
		} else {
			return str;
		}
	}

	private Object mid(Object[] args) {
		int x = toInt(args[1]);
		if (args.length == 2) {
			String str = toString(args[0]);
			return str.substring(x - 1);
		} else if (args[0] instanceof Reference) {
			return new MidReference((Reference) args[0], x, toInt(args[2]));
		} else {
			String str = toString(args[0]);
			return str.substring(x - 1, x - 1 + toInt(args[2]));
		}
	}

	private Integer pos(String str1, String str2) {
		return str1.indexOf(str2) + 1;
	}

	private Integer posB(String str1, String str2) {
		int pos = str1.indexOf(str2);
		if (pos < 0) {
			return 0;
		}
		String left = str1.substring(0, pos);
		return left.getBytes().length + 1;
	}

	private String replace(String str1, String str2, String str3) {
		return str1.replace(str2, str3);
	}

	private Double sum(int l, int t, int r, int b) {
		double sumValue = 0.0;
		Environment env = macro.getEnvironment();
		for (int y = t; y <= b; y++) {
			for (int x = l; x <= r; x++) {
				Object value = env.getValueAt0(y-1, x-1);
				if (value instanceof Double) {
					sumValue += (Double) value;
				}
			}
		}
		return sumValue;
	}

	private Double avr(int l, int t, int r, int b) {
		double sumValue = 0.0;
		int count = 0;
		Environment env = macro.getEnvironment();
		for (int y = t; y <= b; y++) {
			for (int x = l; x <= r; x++) {
				Object value = env.getValueAt0(y-1, x-1);
				if (value instanceof Double) {
					sumValue += (Double) value;
					count++;
				}
			}
		}
		return (sumValue / count);
	}

	private static Random getRandom() {
		if (rand == null) {
			rand = new SecureRandom();
		}
		return rand;
	}

	private static Integer random(int x) {
		Random r = getRandom();
		return r.nextInt(x);
	}

	private Integer getRowHeight(Object[] args) {
		ICassavaViewModel viewModel = macro.getEnvironment().getViewModel();
		if (args.length == 0) {
			return viewModel.getDefaultRowHeight();
		} else {
			return viewModel.getRowHeight(toInt(args[0]));
		}
	}

	private Integer getColWidth(Object[] args) {
		ICassavaViewModel viewModel = macro.getEnvironment().getViewModel();
		if (args.length == 0) {
			return viewModel.getDefaultColWidth();
		} else {
			return viewModel.getColWidth(toInt(args[0]));
		}
	}

	private String getFileName() {
		ICassavaTableModel model = macro.getEnvironment().getModel();
		String fileName = model.getFilename();
		if (fileName == null) {
			return EMPTY;
		}
		int index = Math.max(fileName.lastIndexOf('\\'), fileName
				.lastIndexOf('/'));
		if (index >= 0) {
			return fileName.substring(index + 1);
		} else {
			return fileName;
		}
	}

	private String getFilePath() {
		ICassavaTableModel model = macro.getEnvironment().getModel();
		String fileName = model.getFilename();
		if (fileName == null) {
			return EMPTY;
		}
		int index = Math.max(fileName.lastIndexOf('\\'), fileName
				.lastIndexOf('/'));
		if (index >= 0) {
			return fileName.substring(0, index + 1);
		} else {
			return EMPTY;
		}
	}

	private Integer getYear() {
		return Calendar.getInstance().get(Calendar.YEAR);
	}

	private Integer getMonth() {
		return Calendar.getInstance().get(Calendar.MONTH) + 1;
	}

	private Integer getDate() {
		return Calendar.getInstance().get(Calendar.DATE);
	}

	private Integer getHours() {
		return Calendar.getInstance().get(Calendar.HOUR_OF_DAY);
	}

	private Integer getMinutes() {
		return Calendar.getInstance().get(Calendar.MINUTE);
	}

	private Integer getSeconds() {
		return Calendar.getInstance().get(Calendar.SECOND);
	}
}
