package net.asukaze.cassava.macro.cui;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;

import net.asukaze.cassava.core.model.DataFormat;
import net.asukaze.cassava.core.model.ICassavaTableModel;
import net.asukaze.cassava.io.util.CsvReader;
import net.asukaze.cassava.io.util.CsvWriter;
import net.asukaze.cassava.macro.APIFunctionHandler;
import net.asukaze.cassava.macro.Environment;
import net.asukaze.cassava.macro.FunctionNotFoundException;
import net.asukaze.cassava.macro.Macro;
import net.asukaze.cassava.macro.MacroException;
import net.asukaze.cassava.macro.api.CUIMessageBoxFunctionHandler;
import net.asukaze.cassava.macro.api.CoreFunction;
import net.asukaze.cassava.macro.api.TableWriteFunctionHandler;
import net.asukaze.cassava.macro.parser.ParseException;

public class InteractiveFunctionHandler implements APIFunctionHandler {
	private static final String LIST_CROSS = "+";
	private static final String LIST_HRULE = "-";
	private static final String LIST_VRULE = "|";
	private static final String SPACER = " ";
	private static final String TITLE_PREFIX = "[ "; //$NON-NLS-1$
	private static final String TITLE_POSTFIX = " ]"; //$NON-NLS-1$

	private BufferedReader in;
	private PrintStream out;
	private DataFormat dataFormat;

	public InteractiveFunctionHandler(BufferedReader in, PrintStream out,
			DataFormat dataFormat) {
		this.in = in;
		this.out = out;
		this.dataFormat = dataFormat;
	}

	public Object runAPIFunction(String name, Object[] args, Macro macro)
			throws IOException {
		int n = args.length;
		if ("list".equals(name) && n <= 2) {
			list(args, macro);
			return null;
		} else if ("open".equals(name) && n == 1) {
			open(macro.toString(args[0]), macro);
			return null;
		} else if ("save".equals(name) && n <= 1) {
			if (n == 0) {
				save(null, macro);
			} else {
				save(macro.toString(args[0]), macro);
			}
			return null;
		} else if ("run".equals(name) && n == 1) {
			run(macro.toString(args[0]), macro);
			return null;
		} else if ("ls".equals(name) && n <= 0) {
			if (n == 0) {
				ls(".");
			} else {
				ls(macro.toString(args[0]));
			}
			return null;
		} else if ("pwd".equals(name) && n == 0) {
			return pwd();
		}
		return new FunctionNotFoundException(name, n);
	}

	private void list(Object[] args, Macro macro) throws IOException {
		int top;
		int bottom;
		Environment env = macro.getEnvironment();
		if (args.length == 0) {
			top = 1;
			bottom = env.getModel().getRowCount();
		} else if (args.length == 1) {
			top = macro.toInt(args[0]);
			bottom = top;
		} else {
			top = macro.toInt(args[0]);
			bottom = macro.toInt(args[1]);
		}
		int right = env.getModel().getColumnCount();
		if (bottom < 1) {
			bottom = 1;
		}
		if (right < 1) {
			right = 1;
		}
		list(1, top, right, bottom, macro);
	}

	private void list(int left, int top, int right, int bottom, Macro macro)
			throws IOException {
		Environment env = macro.getEnvironment();
		List<Integer> widths = new ArrayList<Integer>(right - left + 1);
		for (int x = left; x <= right; x++) {
			int max = 1;
			for (int y = top; y <= bottom; y++) {
				int w = macro.toString(env.getValueAt0(y-1, x-1)).length();
				if (w > max) {
					max = w;
				}
			}
			widths.add(max);
		}

		writeRowSeparator(widths);
		for (int y = top; y <= bottom; y++) {
			out.print(LIST_VRULE);
			for (int x = left; x <= right; x++) {
				String value = macro.toString(env.getValueAt0(y-1, x-1));
				int width = widths.get(x - left);
				out.print(value);
				for (int i = value.length(); i < width; i++) {
					out.print(SPACER);
				}
				out.print(LIST_VRULE);
			}
			out.println();
		}
		writeRowSeparator(widths);
	}

	private void writeRowSeparator(List<Integer> widths) {
		out.print(LIST_CROSS);
		for (Integer w : widths) {
			for (int i = 0; i < w; i++) {
				out.print(LIST_HRULE);
			}
			out.print(LIST_CROSS);
		}
		out.println();
	}

	public static boolean confirmIfModified(ICassavaTableModel model,
			String message, BufferedReader in, PrintStream out)
			throws IOException {
		if (model.isModified() && in != null && out != null) {
			out.println(TITLE_PREFIX + "ۑĂȂύX܂B" + TITLE_POSTFIX);
			out.println(message);
			String prompt = "Yes(Y), No(N) ? ";
			while (true) {
				out.print(prompt);
				String line = in.readLine();
				if (line.charAt(0) == 'Y' || line.charAt(0) == 'y') {
					return true;
				} else if (line.charAt(0) == 'N' || line.charAt(0) == 'n') {
					return false;
				}
			}
		}
		return true;
	}

	private void open(String filename, Macro macro) throws IOException {
		ICassavaTableModel model = macro.getEnvironment().getModel();
		if (confirmIfModified(model, "t@CJ܂H", in, out)) {
			FileReader reader = new FileReader(filename); // TODO: R[h
			CsvReader csvReader = new CsvReader(dataFormat, reader);
			try {
				csvReader.readTable(model);
			} finally {
				csvReader.close();
			}
			out.println("\"" + filename + "\" opened.");
			model.setFilename(filename);
		}
	}

	private void save(String filename, Macro macro) throws IOException {
		ICassavaTableModel model = macro.getEnvironment().getModel();
		if (filename == null) {
			filename = model.getFilename();
			if (filename == null) {
				throw new IOException("t@C܂B");
			}
		}
		FileWriter writer = new FileWriter(filename); // TODO: R[hAsR[h
		CsvWriter csvWriter = new CsvWriter(dataFormat, writer, System
				.getProperty("line.separator"));
		try {
			csvWriter.writeTable(model);
			out.println("\"" + filename + "\" saved.");
			model.setModified(false);
			model.setFilename(filename);
		} finally {
			csvWriter.close();
		}
	}

	private static void setupBatchMacro(Macro macro, BufferedReader in,
			PrintStream out) {
		macro.clearAPIFunctionHandlers();
		macro.registerAPIFunctionHandler(new CoreFunction.Handler());
		macro.registerAPIFunctionHandler(new TableWriteFunctionHandler());
		macro.registerAPIFunctionHandler(new CUIMessageBoxFunctionHandler(in,
				out));
		macro.registerAPIFunctionHandler(new MenuEmulator(in, out));
	}

	private void run(String filename, Macro macro) throws IOException {
		List<APIFunctionHandler> handlers = macro.getAPIFunctionHandlers();
		try {
			setupBatchMacro(macro, in, out);
			FileReader reader = new FileReader(filename);
			macro.run(reader);
		} catch (ParseException ex) {
			throw new MacroException(ex);
		} finally {
			macro.setAPIFunctionHandlers(handlers);
		}
	}

	private void ls(String path) {
		String[] list = new File(path).list(new FilenameFilter() {
			public boolean accept(File dir, String name) {
				return !(name.startsWith("."));
			}
		});
		int max = 1;
		for (String file : list) {
			if (file.length() > max) {
				max = file.length();
			}
		}
		int width = Math.min(80, max + 2);
		int count = list.length;
		int cols = 80 / width;
		int rows = (int) (Math.ceil((double) count / (double) cols));
		for (int y = 0; y < rows; y++) {
			for (int i = y; i < count; i += rows) {
				String file = list[i];
				out.print(file);
				if (i + rows < count) {
					for (int j = file.length(); j < width; j++) {
						out.print(SPACER);
					}
				}
			}
			out.println();
		}
	}

	private String pwd() {
		return new File(".").getAbsoluteFile().getParent();
	}

	public void close() {
	}
}
